# "title/wad.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/WAD_files for details about the WAD format

import io
import binascii
from ..shared import _align_value, _pad_bytes


class WAD:
    """
    A WAD object that allows for either loading and editing an existing WAD or creating a new WAD from raw data.

    Attributes
    ----------
    wad_type : str
        The type of WAD, either ib for boot2 or Is for normal installable WADs.
    wad_cert_size : int
        The size of the WAD's certificate.
    wad_crl_size : int
        The size of the WAD's crl.
    wad_tik_size : int
        The size of the WAD's Ticket.
    wad_tmd_size : int
        The size of the WAD's TMD.
    wad_content_size : int
        The size of WAD's total content region.
    wad_meta_size : int
        The size of the WAD's meta/footer.
    """
    def __init__(self):
        self.wad_hdr_size: int = 64
        self.wad_type: str = "Is"
        self.wad_version: bytes = b'\x00\x00'
        # === Sizes ===
        self.wad_cert_size: int = 0
        self.wad_crl_size: int = 0
        self.wad_tik_size: int = 0
        self.wad_tmd_size: int = 0
        # This is the size of the content region, which contains all app files combined.
        self.wad_content_size: int = 0
        self.wad_meta_size: int = 0
        # === Data ===
        self.wad_cert_data: bytes = b''
        self.wad_crl_data: bytes = b''
        self.wad_tik_data: bytes = b''
        self.wad_tmd_data: bytes = b''
        self.wad_content_data: bytes = b''
        self.wad_meta_data: bytes = b''

    def load(self, wad_data: bytes) -> None:
        """
        Loads raw WAD data and sets all attributes of the WAD object. This allows for manipulating an already
        existing WAD file.

        Parameters
        ----------
        wad_data : bytes
            The data for the WAD file to load.
        """
        with io.BytesIO(wad_data) as wad_data:
            # Read the first 8 bytes of the file to ensure that it's a WAD. Has two possible valid values for the two
            # different types of WADs that might be encountered.
            wad_data.seek(0x0)
            wad_magic_bin = wad_data.read(8)
            wad_magic_hex = binascii.hexlify(wad_magic_bin)
            wad_magic = str(wad_magic_hex.decode())
            if wad_magic != "0000002049730000" and wad_magic != "0000002069620000":
                raise TypeError("This is not a valid WAD file!")
            # ====================================================================================
            # Get the sizes of each data region contained within the WAD.
            # ====================================================================================
            # Header length, which will always be 64 bytes, as it is padded out if it is shorter.
            self.wad_hdr_size = 64
            # WAD type, denoting whether this WAD contains boot2 ("ib"), or anything else ("Is").
            wad_data.seek(0x04)
            self.wad_type = str(wad_data.read(2).decode())
            # WAD version, this is always 0.
            wad_data.seek(0x06)
            self.wad_version = wad_data.read(2)
            # WAD cert size.
            wad_data.seek(0x08)
            self.wad_cert_size = int(binascii.hexlify(wad_data.read(4)), 16)
            # WAD crl size.
            wad_data.seek(0x0c)
            self.wad_crl_size = int(binascii.hexlify(wad_data.read(4)), 16)
            # WAD ticket size.
            wad_data.seek(0x10)
            self.wad_tik_size = int(binascii.hexlify(wad_data.read(4)), 16)
            # WAD TMD size.
            wad_data.seek(0x14)
            self.wad_tmd_size = int(binascii.hexlify(wad_data.read(4)), 16)
            # WAD content size. This needs to be rounded now, because with some titles (primarily IOS?), there can be
            # extra bytes past the listed end of the content that is needed for decryption.
            wad_data.seek(0x18)
            self.wad_content_size = int(binascii.hexlify(wad_data.read(4)), 16)
            self.wad_content_size = _align_value(self.wad_content_size, 16)
            # Time/build stamp for the title contained in the WAD.
            wad_data.seek(0x1c)
            self.wad_meta_size = int(binascii.hexlify(wad_data.read(4)), 16)
            # ====================================================================================
            # Calculate file offsets from sizes. Every section of the WAD is padded out to a multiple of 0x40.
            # ====================================================================================
            wad_cert_offset = self.wad_hdr_size
            # crl isn't ever used, however an entry for its size exists in the header, so it's calculated just in case.
            wad_crl_offset = _align_value(wad_cert_offset + self.wad_cert_size)
            wad_tik_offset = _align_value(wad_crl_offset + self.wad_crl_size)
            wad_tmd_offset = _align_value(wad_tik_offset + self.wad_tik_size)
            wad_content_offset = _align_value(wad_tmd_offset + self.wad_tmd_size)
            # meta isn't guaranteed to be used, but some older SDK titles use it, and not reading it breaks things.
            wad_meta_offset = _align_value(wad_content_offset + self.wad_content_size)
            # ====================================================================================
            # Load data for each WAD section based on the previously calculated offsets.
            # ====================================================================================
            # Cert data.
            wad_data.seek(wad_cert_offset)
            self.wad_cert_data = wad_data.read(self.wad_cert_size)
            # Crl data.
            wad_data.seek(wad_crl_offset)
            self.wad_crl_data = wad_data.read(self.wad_crl_size)
            # Ticket data.
            wad_data.seek(wad_tik_offset)
            self.wad_tik_data = wad_data.read(self.wad_tik_size)
            # TMD data.
            wad_data.seek(wad_tmd_offset)
            self.wad_tmd_data = wad_data.read(self.wad_tmd_size)
            # Content data.
            wad_data.seek(wad_content_offset)
            self.wad_content_data = wad_data.read(self.wad_content_size)
            # Meta data.
            wad_data.seek(wad_meta_offset)
            self.wad_meta_data = wad_data.read(self.wad_meta_size)

    def dump(self) -> bytes:
        """
        Dumps the WAD object into the raw WAD file. This allows for creating a WAD file from the data contained in
        the WAD object.

        Returns
        -------
        bytes
            The full WAD file as bytes.
        """
        wad_data = b''
        # Lead-in data.
        wad_data += b'\x00\x00\x00\x20'
        # WAD type.
        wad_data += str.encode(self.wad_type)
        # WAD version.
        wad_data += self.wad_version
        # WAD cert size.
        wad_data += int.to_bytes(self.wad_cert_size, 4)
        # WAD crl size.
        wad_data += int.to_bytes(self.wad_crl_size, 4)
        # WAD ticket size.
        wad_data += int.to_bytes(self.wad_tik_size, 4)
        # WAD TMD size.
        wad_data += int.to_bytes(self.wad_tmd_size, 4)
        # WAD content size.
        wad_data += int.to_bytes(self.wad_content_size, 4)
        # WAD meta size.
        wad_data += int.to_bytes(self.wad_meta_size, 4)
        wad_data = _pad_bytes(wad_data)
        # Retrieve the cert data and write it out.
        wad_data += self.get_cert_data()
        wad_data = _pad_bytes(wad_data)
        # Retrieve the crl data and write it out.
        wad_data += self.get_crl_data()
        wad_data = _pad_bytes(wad_data)
        # Retrieve the ticket data and write it out.
        wad_data += self.get_ticket_data()
        wad_data = _pad_bytes(wad_data)
        # Retrieve the TMD data and write it out.
        wad_data += self.get_tmd_data()
        wad_data = _pad_bytes(wad_data)
        # Retrieve the content data and write it out.
        wad_data += self.get_content_data()
        wad_data = _pad_bytes(wad_data)
        # Retrieve the meta/footer data and write it out.
        wad_data += self.get_meta_data()
        wad_data = _pad_bytes(wad_data)
        return wad_data

    def get_wad_type(self) -> str:
        """
        Gets the type of the WAD.

        Returns
        -------
        str
            The type of the WAD. This is 'Is', unless the WAD contains boot2, where it is 'ib'.
        """
        return self.wad_type

    def get_cert_data(self) -> bytes:
        """
        Gets the certificate data from the WAD.

        Returns
        -------
        bytes
            The certificate data.
        """
        return self.wad_cert_data

    def get_crl_data(self) -> bytes:
        """
        Gets the crl data from the WAD, if it exists.

        Returns
        -------
        bytes
            The crl data.
        """
        return self.wad_crl_data

    def get_ticket_data(self) -> bytes:
        """
        Gets the ticket data from the WAD.

        Returns
        -------
        bytes
            The ticket data.
        """
        return self.wad_tik_data

    def get_tmd_data(self) -> bytes:
        """
        Returns the TMD data from the WAD.

        Returns
        -------
        bytes
            The TMD data.
        """
        return self.wad_tmd_data

    def get_content_data(self) -> bytes:
        """
        Gets the content of the WAD.

        Returns
        -------
        bytes
            The content data.
        """
        return self.wad_content_data

    def get_meta_data(self) -> bytes:
        """
        Gets the meta region of the WAD, which is typically unused.

        Returns
        -------
        bytes
            The meta region.
        """
        return self.wad_meta_data

    def set_cert_data(self, cert_data) -> None:
        """
        Sets the certificate data of the WAD. Also calculates the new size.

        Parameters
        ----------
        cert_data : bytes
            The new certificate data.
        """
        self.wad_cert_data = cert_data
        # Calculate the size of the new cert data.
        self.wad_cert_size = len(cert_data)

    def set_crl_data(self, crl_data) -> None:
        """
        Sets the crl data of the WAD. Also calculates the new size.

        Parameters
        ----------
        crl_data : bytes
            The new crl data.
        """
        self.wad_crl_data = crl_data
        # Calculate the size of the new crl data.
        self.wad_crl_size = len(crl_data)

    def set_tmd_data(self, tmd_data) -> None:
        """
        Sets the TMD data of the WAD. Also calculates the new size.

        Parameters
        ----------
        tmd_data : bytes
            The new TMD data.
        """
        self.wad_tmd_data = tmd_data
        # Calculate the size of the new TMD data.
        self.wad_tmd_size = len(tmd_data)

    def set_ticket_data(self, tik_data) -> None:
        """
        Sets the Ticket data of the WAD. Also calculates the new size.

        Parameters
        ----------
        tik_data : bytes
            The new TMD data.
        """
        self.wad_tik_data = tik_data
        # Calculate the size of the new Ticket data.
        self.wad_tik_size = len(tik_data)

    def set_content_data(self, content_data, size: int = None) -> None:
        """
        Sets the content data of the WAD. Also calculates the new size.

        Parameters
        ----------
        content_data : bytes
            The new content data.
        size : int, option
            The size of the new content data.
        """
        self.wad_content_data = content_data
        # Calculate the size of the new content data, if one wasn't supplied.
        if size is None:
            self.wad_content_size = len(content_data)
        else:
            self.wad_content_size = size

    def set_meta_data(self, meta_data) -> None:
        """
        Sets the meta data of the WAD. Also calculates the new size.

        Parameters
        ----------
        meta_data : bytes
            The new meta data.
        """
        self.wad_meta_data = meta_data
        # Calculate the size of the new meta data.
        self.wad_meta_size = len(meta_data)



# "title/tmd.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Title_metadata for details about the TMD format

import io
import binascii
import hashlib
import math
import struct
from typing import List
from enum import IntEnum as _IntEnum
from ..types import _ContentRecord
from ..shared import _bitmask
from .util import title_ver_dec_to_standard, title_ver_standard_to_dec


class TMD:
    """
    A TMD object that allows for either loading and editing an existing TMD or creating one manually if desired.

    Attributes
    ----------
    title_id : str
        The title ID of the title listed in the TMD.
    title_version : int
        The version of the title listed in the TMD.
    tmd_version : int
        The version of the TMD.
    ios_tid : str
        The title ID of the IOS the title runs on.
    ios_version : int
        The IOS version the title runs on.
    num_contents : int
        The number of contents listed in the TMD.
    """
    def __init__(self):
        self.blob_header: bytes = b''
        self.signature_type: int = 0
        self.signature: bytes = b''
        self.signature_issuer: str = ""  # Follows the format "Root-CA%08x-CP%08x"
        self.tmd_version: int = 0  # This seems to always be 0 no matter what?
        self.ca_crl_version: int = 0  # Certificate Authority Certificate Revocation List version
        self.signer_crl_version: int = 0  # Certificate Policy Certificate Revocation List version
        self.vwii: int = 0  # Whether the title is for the vWii. 0 = No, 1 = Yes
        self.ios_tid: str = ""  # The Title ID of the IOS version the associated title runs on.
        self.ios_version: int = 0  # The IOS version the associated title runs on.
        self.title_id: str = ""  # The Title ID of the associated title.
        self.title_type: bytes = b''  # The type of the associated title. Should always be 00000001 in a Wii TMD.
        self.group_id: int = 0  # The ID of the publisher of the associated title.
        self.region: int = 0  # The ID of the region of the associated title.
        self.ratings: bytes = b''  # The parental controls rating of the associated title.
        self.reserved1: bytes = b''  # Unknown data labeled "Reserved" on WiiBrew.
        self.ipc_mask: bytes = b''
        self.reserved2: bytes = b''  # Other "Reserved" data from WiiBrew.
        self.access_rights: int = 0
        self.title_version: int = 0  # The version of the associated title.
        self.title_version_converted: int = 0  # The title version in vX.X format.
        self.num_contents: int = 0  # The number of contents contained in the associated title.
        self.boot_index: int = 0  # The content index that contains the bootable executable.
        self.minor_version: int = 0  # Minor version (unused typically).
        self.content_records: List[_ContentRecord] = []

    def load(self, tmd: bytes) -> None:
        """
        Loads raw TMD data and sets all attributes of the TMD object. This allows for manipulating an already
        existing TMD.

        Parameters
        ----------
        tmd : bytes
            The data for the TMD you wish to load.
        """
        with io.BytesIO(tmd) as tmd_data:
            # ====================================================================================
            # Parses each of the keys contained in the TMD.
            # ====================================================================================
            # Signature type.
            tmd_data.seek(0x0)
            self.signature_type = tmd_data.read(4)
            # Signature data.
            tmd_data.seek(0x04)
            self.signature = tmd_data.read(256)
            # Signing certificate issuer.
            tmd_data.seek(0x140)
            self.signature_issuer = str(tmd_data.read(64).replace(b'\x00', b'').decode())
            # TMD version, seems to usually be 0, but I've seen references to other numbers.
            tmd_data.seek(0x180)
            self.tmd_version = int.from_bytes(tmd_data.read(1))
            # Certificate Authority CRL version.
            tmd_data.seek(0x181)
            self.ca_crl_version = int.from_bytes(tmd_data.read(1))
            # Certificate Policy CRL version.
            tmd_data.seek(0x182)
            self.signer_crl_version = int.from_bytes(tmd_data.read(1))
            # If this is a vWii title or not.
            tmd_data.seek(0x183)
            self.vwii = int.from_bytes(tmd_data.read(1))
            # TID of the IOS to use for the title, set to 0 if this title is the IOS, set to boot2 version if boot2.
            tmd_data.seek(0x184)
            ios_version_bin = tmd_data.read(8)
            ios_version_hex = binascii.hexlify(ios_version_bin)
            self.ios_tid = str(ios_version_hex.decode())
            # Get IOS version based on TID.
            self.ios_version = int(self.ios_tid[-2:], 16)
            # Title ID of the title.
            tmd_data.seek(0x18C)
            title_id_bin = tmd_data.read(8)
            title_id_hex = binascii.hexlify(title_id_bin)
            self.title_id = str(title_id_hex.decode())
            # Type of the title. This is an internal property used to show if this title is for the ill-fated
            # NetCard (0), or the Wii (1), and is therefore always 1 for Wii TMDs.
            tmd_data.seek(0x194)
            self.title_type = tmd_data.read(4)
            # Publisher of the title.
            tmd_data.seek(0x198)
            self.group_id = int.from_bytes(tmd_data.read(2))
            # Region of the title, 0 = JAP, 1 = USA, 2 = EUR, 3 = WORLD, 4 = KOR.
            tmd_data.seek(0x19C)
            region_hex = tmd_data.read(2)
            self.region = int.from_bytes(region_hex)
            # Content rating of the title for parental controls. Likely based on ESRB, CERO, PEGI, etc. rating.
            tmd_data.seek(0x19E)
            self.ratings = tmd_data.read(16)
            # "Reserved" data 1.
            tmd_data.seek(0x1AE)
            self.reserved1 = tmd_data.read(12)
            # IPC mask.
            tmd_data.seek(0x1BA)
            self.ipc_mask = tmd_data.read(12)
            # "Reserved" data 2.
            tmd_data.seek(0x1C6)
            self.reserved2 = tmd_data.read(18)
            # Access rights of the title; DVD-video and AHB access.
            tmd_data.seek(0x1D8)
            self.access_rights = int.from_bytes(tmd_data.read(4))
            # Version number straight from the TMD.
            tmd_data.seek(0x1DC)
            self.title_version = int.from_bytes(tmd_data.read(2))
            # Calculate the converted version number via util module.
            self.title_version_converted = title_ver_dec_to_standard(self.title_version, self.title_id, bool(self.vwii))
            # The number of contents listed in the TMD.
            tmd_data.seek(0x1DE)
            self.num_contents = int.from_bytes(tmd_data.read(2))
            # The content index that contains the bootable executable.
            tmd_data.seek(0x1E0)
            self.boot_index = int.from_bytes(tmd_data.read(2))
            # The minor version of the title (typically unused).
            tmd_data.seek(0x1E2)
            self.minor_version = int.from_bytes(tmd_data.read(2))
            # Get content records for the number of contents in num_contents.
            self.content_records = []
            for content in range(0, self.num_contents):
                tmd_data.seek(0x1E4 + (36 * content))
                content_record_hdr = struct.unpack(">LHH4x4s20s", tmd_data.read(36))
                self.content_records.append(
                    _ContentRecord(int(content_record_hdr[0]), int(content_record_hdr[1]),
                                   int(content_record_hdr[2]), int.from_bytes(content_record_hdr[3]),
                                   binascii.hexlify(content_record_hdr[4])))

    def dump(self) -> bytes:
        """
        Dumps the TMD object back into bytes.

        Returns
        -------
        bytes
            The full TMD file as bytes.
        """
        tmd_data = b''
        # Signature type.
        tmd_data += self.signature_type
        # Signature data.
        tmd_data += self.signature
        # Padding to 64 bytes.
        tmd_data += b'\x00' * 60
        # Signing certificate issuer.
        signature_issuer = self.signature_issuer.encode()
        while len(signature_issuer) < 0x40:
            signature_issuer += b'\x00'
        tmd_data += signature_issuer
        # TMD version.
        tmd_data += int.to_bytes(self.tmd_version, 1)
        # Certificate Authority CRL version.
        tmd_data += int.to_bytes(self.ca_crl_version, 1)
        # Certificate Policy CRL version.
        tmd_data += int.to_bytes(self.signer_crl_version, 1)
        # If this is a vWii title or not.
        tmd_data += int.to_bytes(self.vwii, 1)
        # IOS Title ID.
        tmd_data += binascii.unhexlify(self.ios_tid)
        # Title's Title ID.
        tmd_data += binascii.unhexlify(self.title_id)
        # Title type.
        tmd_data += self.title_type
        # Group ID.
        tmd_data += int.to_bytes(self.group_id, 2)
        # 2 bytes of zero for reasons.
        tmd_data += b'\x00\x00'
        # Region.
        tmd_data += int.to_bytes(self.region, 2)
        # Parental Controls Ratings.
        tmd_data += self.ratings
        # "Reserved" 1.
        tmd_data += self.reserved1
        # IPC mask.
        tmd_data += self.ipc_mask
        # "Reserved" 2.
        tmd_data += self.reserved2
        # Access rights.
        tmd_data += int.to_bytes(self.access_rights, 4)
        # Title version.
        tmd_data += int.to_bytes(self.title_version, 2)
        # Number of contents.
        tmd_data += int.to_bytes(self.num_contents, 2)
        # Boot index.
        tmd_data += int.to_bytes(self.boot_index, 2)
        # Minor version.
        tmd_data += int.to_bytes(self.minor_version, 2)
        # Iterate over content records, write them back into raw data, then add them to the TMD.
        for content_record in range(self.num_contents):
            content_data = b''
            # Write all fields from the content record.
            content_data += int.to_bytes(self.content_records[content_record].content_id, 4)
            content_data += int.to_bytes(self.content_records[content_record].index, 2)
            content_data += int.to_bytes(self.content_records[content_record].content_type, 2)
            content_data += int.to_bytes(self.content_records[content_record].content_size, 8)
            content_data += binascii.unhexlify(self.content_records[content_record].content_hash)
            # Write the record to the TMD.
            tmd_data += content_data
        return tmd_data

    def fakesign(self) -> None:
        """
        Fakesigns this TMD for the trucha bug.

        This is done by brute-forcing a TMD body hash starting with 00, causing it to pass signature verification on
        older IOS versions that incorrectly check the hash using strcmp() instead of memcmp(). The signature will also
        be erased and replaced with all NULL bytes.

        The hash is brute-forced by incrementing an unused 16-bit integer in the TMD by 1 until an appropriate hash is
        found.

        This modifies the TMD object in place. You will need to call this method after any changes, and before dumping
        the TMD object back into bytes.
        """
        # Clear the signature, so that the hash derived from it is guaranteed to always be
        # '0000000000000000000000000000000000000000'.
        self.signature = b'\x00' * 256
        current_int = 0
        test_hash = ''
        while test_hash[:2] != '00':
            current_int += 1
            self.minor_version = current_int
            # Trim off the first 320 bytes, because we're only looking for the hash of the TMD's body.
            # This is a try-except because an OverflowError will be thrown if the number being used to brute-force the
            # hash gets too big, as it is only a 16-bit integer. If that happens, then fakesigning has failed.
            try:
                test_hash = hashlib.sha1(self.dump()[320:]).hexdigest()
            except OverflowError:
                raise Exception("An error occurred during fakesigning. TMD could not be fakesigned!")

    def get_is_fakesigned(self) -> bool:
        """
        Checks the TMD object to see if it is currently fakesigned. For a description of fakesigning, refer to the
        fakesign() method.

        Returns
        -------
        bool:
            True if the TMD is fakesigned, False otherwise.

        See Also
        --------
        libWiiPy.title.tmd.TMD.fakesign()
        """
        if self.signature != b'\x00' * 256:
            return False
        test_hash = hashlib.sha1(self.dump()[320:]).hexdigest()
        if test_hash[:2] != '00':
            return False
        return True

    def get_title_region(self) -> str:
        """
        Gets the system region specified in the TMD. This is not necessarily the true region of the title, but is the
        hardware region that this title is designed and allowed to be run on.

        Can be one of several possible values:
        'JPN', 'USA', 'EUR', 'None', or 'KOR'.

        Returns
        -------
        str
            The region of the title.
        """
        match self.region:
            case 0:
                return "JPN"
            case 1:
                return "USA"
            case 2:
                return "EUR"
            case 3:
                return "None"
            case 4:
                return "KOR"

    def get_title_type(self) -> str:
        """
        Gets the type of the title this TMD describes. The title_type field is not related to these types.

        Can be one of several possible values:
        'System', 'Game', 'Channel', 'SystemChannel', 'GameChannel', or 'HiddenChannel'

        Returns
        -------
        str
            The type of the title.
        """
        match self.title_id[:8]:
            case '00000001':
                return "System"
            case '00010000':
                return "Game"
            case '00010001':
                return "Channel"
            case '00010002':
                return "SystemChannel"
            case '00010004':
                return "GameChannel"
            case '00010005':
                return "DLC"
            case '00010008':
                return "HiddenChannel"
            case _:
                return "Unknown"

    def get_content_type(self, content_index: int) -> str:
        """
        Gets the type of content contained in the TMD's associated title.

        Can be one of several possible values:
        'Normal', 'Development/Unknown', 'Hash Tree', 'DLC', or 'Shared'

        Parameters
        ----------
        content_index : int
            The index of the content you want the type of.

        Returns
        -------
        str
            The type of content.
        """
        # Get a list of the current content indices, so we can make sure the target one exists. Doing it this way
        # ensures we can find the target, even if the highest content index is greater than the highest literal index.
        current_indices = []
        for record in self.content_records:
            current_indices.append(record.index)
        # This is the literal index in the list of content that we're going to get.
        target_index = current_indices.index(content_index)
        match self.content_records[target_index].content_type:
            case 1:
                return "Normal"
            case 2:
                return "Development/Unknown"
            case 3:
                return "Hash Tree"
            case 16385:
                return "DLC"
            case 32769:
                return "Shared"
            case _:
                return "Unknown"

    def get_content_record(self, record) -> _ContentRecord:
        """
        Gets the content record at the specified index.

        Parameters
        ----------
        record : int
            The content record to be retrieved.

        Returns
        -------
        _ContentRecord
            A ContentRecord object containing the data in the content record.
        """
        if record < self.num_contents:
            return self.content_records[record]
        else:
            raise IndexError("Invalid content record! TMD lists '" + str(self.num_contents - 1) +
                             "' contents but index was '" + str(record) + "'!")

    def get_content_size(self, absolute=False, dlc=False) -> int:
        """
        Gets the installed size of the content listed in the TMD, in bytes. This does not include the size of hash tree
        content, so the size of disc titles will not be calculated. The "absolute" option determines whether shared
        content sizes should be included in the total size or not. This option defaults to False. The "dlc" option
        determines whether DLC content sizes should be included in the total size or not. This option also defaults to
        False.

        Parameters
        ----------
        absolute: bool, optional
            Whether shared contents should be included in the total size or not. Defaults to False.
        dlc: bool, optional
            Whether DLC contents should be included in the total size or not. Defaults to False.

        Returns
        -------
        int
            The installed size of the content, in bytes.
        """
        title_size = 0
        for record in self.content_records:
            if record.content_type == 0x8001:
                if absolute:
                    title_size += record.content_size
            elif record.content_type == 0x4001:
                if dlc:
                    title_size += record.content_size
            elif record.content_type != 3:
                title_size += record.content_size
        return title_size

    def get_content_size_blocks(self, absolute=False, dlc=False) -> int:
        """
        Gets the installed size of the content listed in the TMD, in the Wii's displayed "blocks" format. The
        "absolute" option determines whether shared content sizes should be included in the total size or not. This
        option defaults to False. The "dlc" option determines whether DLC content sizes should be included in the total
        size or not. This option also defaults to False.

        1 Wii block is equal to 128KiB, and if any amount of a block is used, the entire block is considered used.

        Parameters
        ----------
        absolute : bool, optional
            Whether shared contents should be included in the total size or not. Defaults to False.
        dlc: bool, optional
            Whether DLC contents should be included in the total size or not. Defaults to False.

        Returns
        -------
        int
            The installed size of the content, in blocks.
        """
        title_size_bytes = self.get_content_size(absolute, dlc)
        blocks = math.ceil(title_size_bytes / 131072)
        return blocks

    class AccessFlags(_IntEnum):
        AHB = 0
        DVD_VIDEO = 1

    def get_access_right(self, flag: int) -> bool:
        """
        Gets whether the specified access rights flag is enabled or not. This is done by checking the specified bit.
        Possible flags and their corresponding bits are defined in AccessFlags.

        Parameters
        ----------
        flag : int
            The flag to check.

        Returns
        -------
        bool
            True if the flag is enabled, False otherwise.

        See Also
        --------
        libWiiPy.title.tmd.TMD.AccessFlags
        """
        return bool(self.access_rights & _bitmask(flag))

    def set_title_id(self, title_id) -> None:
        """
        Sets the Title ID property of the TMD. Recommended over setting the property directly because of input
        validation.

        Parameters
        ----------
        title_id : str
            The new Title ID of the title.
        """
        if len(title_id) != 16:
            raise ValueError("Invalid Title ID! Title IDs must be 8 bytes long.")
        self.title_id = title_id

    def set_title_version(self, new_version: str | int) -> None:
        """
        Sets the version of the title in the TMD. Recommended over setting the data directly because of input
        validation.

        Accepts either standard form (vX.X) as a string or decimal form (vXXX) as an integer.

        Parameters
        ----------
        new_version : str, int
            The new version of the title. See description for valid formats.
        """
        if type(new_version) is str:
            # Validate string input is in the correct format, then validate that the version isn't higher than v255.0.
            # If checks pass, set that as the converted version, then convert to decimal form and set that as well.
            version_str_split = new_version.split(".")
            if len(version_str_split) != 2:
                raise ValueError("Title version is not valid! String version must be entered in format \"X.X\".")
            if int(version_str_split[0]) > 255 or int(version_str_split[1]) > 255:
                raise ValueError("Title version is not valid! String version number cannot exceed v255.255.")
            self.title_version_converted = new_version
            version_converted = title_ver_standard_to_dec(new_version, self.title_id)
            self.title_version = version_converted
        elif type(new_version) is int:
            # Validate that the version isn't higher than v65280. If the check passes, set that as the title version,
            # then convert to standard form and set that as well.
            if new_version > 65535:
                raise ValueError("Title version is not valid! Integer version number cannot exceed v65535.")
            self.title_version = new_version
            version_converted = title_ver_dec_to_standard(new_version, self.title_id, bool(self.vwii))
            self.title_version_converted = version_converted
        else:
            raise TypeError("Title version type is not valid! Type must be either integer or string.")



# "title/title.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Title for details about how titles are formatted

import math
from .cert import (CertificateChain as _CertificateChain,
                   verify_ca_cert as _verify_ca_cert, verify_cert_sig as _verify_cert_sig,
                   verify_tmd_sig as _verify_tmd_sig, verify_ticket_sig as _verify_ticket_sig)
from .content import ContentRegion as _ContentRegion
from .ticket import Ticket as _Ticket
from .tmd import TMD as _TMD
from .wad import WAD as _WAD
from .crypto import encrypt_title_key


class Title:
    """
    A Title object that contains all components of a title, and allows altering them. Provides higher-level access
    than manually creating WAD, TMD, Ticket, and ContentRegion objects and ensures that any data that needs to match
    between files matches.

    Attributes
    ----------
    wad: WAD
        A WAD object of a WAD containing the title's data.
    cert_chain: CertificateChain
        The chain of certificates used to verify the contents of a title.
    tmd: TMD
        A TMD object of the title's TMD.
    ticket: Ticket
        A Ticket object of the title's Ticket.
    content: ContentRegion
        A ContentRegion object containing the title's contents.
    """
    def __init__(self):
        self.wad: _WAD = _WAD()
        self.cert_chain: _CertificateChain = _CertificateChain()
        self.tmd: _TMD = _TMD()
        self.ticket: _Ticket = _Ticket()
        self.content: _ContentRegion = _ContentRegion()

    def load_wad(self, wad: bytes) -> None:
        """
        Load existing WAD data into the title and create WAD, TMD, Ticket, and ContentRegion objects based off of it
        to allow you to modify that data. Note that this will overwrite any existing data for this title.

        Parameters
        ----------
        wad : bytes
            The data for the WAD you wish to load.
        """
        # Create a new WAD object based on the WAD data provided.
        self.wad = _WAD()
        self.wad.load(wad)
        # Load the certificate chain.
        self.cert_chain = _CertificateChain()
        self.cert_chain.load(self.wad.get_cert_data())
        # Load the TMD.
        self.tmd = _TMD()
        self.tmd.load(self.wad.get_tmd_data())
        # Load the ticket.
        self.ticket = _Ticket()
        self.ticket.load(self.wad.get_ticket_data())
        # Load the content.
        self.content = _ContentRegion()
        self.content.load(self.wad.get_content_data(), self.tmd.content_records)
        # Ensure that the Title IDs of the TMD and Ticket match before doing anything else. If they don't, throw an
        # error because clearly something strange has gone on with the WAD and editing it probably won't work.
        #if self.tmd.title_id != str(self.ticket.title_id.decode()):
        #    raise ValueError("The Title IDs of the TMD and Ticket in this WAD do not match. This WAD appears to be "
        #                     "invalid.")

    def dump_wad(self) -> bytes:
        """
        Dumps all title components (TMD, Ticket, and contents) back into the WAD object, and then dumps the WAD back
        into raw data and returns it.

        Returns
        -------
        wad_data : bytes
            The raw data of the WAD.
        """
        # Set WAD type to ib if the title being packed is boot2.
        if self.tmd.title_id == "0000000100000001":
            self.wad.wad_type = "ib"
        # Dump the certificate chain and set it in the WAD.
        self.wad.set_cert_data(self.cert_chain.dump())
        # Dump the TMD and set it in the WAD.
        # This requires updating the content records and number of contents in the TMD first.
        self.tmd.content_records = self.content.content_records  # This may not be needed because it's a ref already
        self.tmd.num_contents = len(self.content.content_records)
        self.wad.set_tmd_data(self.tmd.dump())
        # Dump the Ticket and set it in the WAD.
        self.wad.set_ticket_data(self.ticket.dump())
        # Dump the ContentRegion and set it in the WAD.
        content_data, content_size = self.content.dump()
        self.wad.set_content_data(content_data, content_size)
        return self.wad.dump()

    def load_cert_chain(self, cert_chain: bytes) -> None:
        """
        Load an existing certificate chain into the title. Note that this will overwrite any existing certificate chain
        data for this title.

        Parameters
        ----------
        cert_chain: bytes
            The data for the certificate chain to load.
        """
        self.cert_chain.load(cert_chain)


    def load_tmd(self, tmd: bytes) -> None:
        """
        Load existing TMD data into the title. Note that this will overwrite any existing TMD data for this title.

        Parameters
        ----------
        tmd : bytes
            The data for the TMD to load.
        """
        self.tmd.load(tmd)

    def load_ticket(self, ticket: bytes) -> None:
        """
        Load existing Ticket data into the title. Note that this will overwrite any existing Ticket data for this
        title.

        Parameters
        ----------
        ticket : bytes
            The data for the Ticket to load.
        """
        self.ticket.load(ticket)

    def load_content_records(self) -> None:
        """
        Load content records from the TMD into the ContentRegion to allow loading content files based on the records.
        This requires that a TMD has already been loaded and will throw an exception if it isn't.
        """
        if not self.tmd.content_records:
            ValueError("No TMD appears to have been loaded, so content records cannot be read from it.")
        # Load the content records into the ContentRegion object, and update the number of contents.
        self.content.content_records = self.tmd.content_records
        self.content.num_contents = self.tmd.num_contents

    def set_title_id(self, title_id: str) -> None:
        """
        Sets the Title ID of the title in both the TMD and Ticket. This also re-encrypts the Title Key as the Title Key
        is used as the IV for decrypting it.

        Parameters
        ----------
        title_id : str
            The new Title ID of the title.
        """
        if len(title_id) != 16:
            raise ValueError("Invalid Title ID! Title IDs must be 8 bytes long.")
        self.tmd.set_title_id(title_id)
        title_key_decrypted = self.ticket.get_title_key()
        self.ticket.set_title_id(title_id)
        title_key_encrypted = encrypt_title_key(title_key_decrypted, self.ticket.common_key_index, title_id,
                                                self.ticket.is_dev)
        self.ticket.title_key_enc = title_key_encrypted

    def set_title_version(self, title_version: str | int) -> None:
        """
        Sets the version of the title in both the TMD and Ticket.

        Accepts either standard form (vX.X) as a string or decimal form (vXXX) as an integer.

        Parameters
        ----------
        title_version : str, int
            The new version of the title. See description for valid formats.
        """
        self.tmd.set_title_version(title_version)
        self.ticket.set_title_version(title_version)

    def get_content_by_index(self, index: id, skip_hash=False) -> bytes:
        """
        Gets an individual content from the content region based on the provided index, in decrypted form.

        Parameters
        ----------
        index : int
            The index of the content you want to get.
        skip_hash : bool, optional
            Skip the hash check and return the content regardless of its hash. Defaults to false.

        Returns
        -------
        bytes
            The decrypted content listed in the content record.
        """
        dec_content = self.content.get_content_by_index(index, self.ticket.get_title_key(), skip_hash)
        return dec_content

    def get_content_by_cid(self, cid: int, skip_hash=False) -> bytes:
        """
        Gets an individual content from the content region based on the provided Content ID, in decrypted form.

        Parameters
        ----------
        cid : int
            The Content ID of the content you want to get. Expected to be in decimal form.
        skip_hash : bool, optional
            Skip the hash check and return the content regardless of its hash. Defaults to false.

        Returns
        -------
        bytes
            The decrypted content listed in the content record.
        """
        dec_content = self.content.get_content_by_cid(cid, self.ticket.get_title_key(), skip_hash)
        return dec_content

    def get_title_size(self, absolute=False) -> int:
        """
        Gets the installed size of the title, including the TMD and Ticket, in bytes. The "absolute" option determines
        whether shared content sizes should be included in the total size or not. This option defaults to False.

        Parameters
        ----------
        absolute : bool, optional
            Whether shared contents should be included in the total size or not. Defaults to False.

        Returns
        -------
        int
            The installed size of the title, in bytes.
        """
        title_size = 0
        # Dumping and measuring the TMD and Ticket this way to ensure that any changes to them are measured properly.
        # Yes, the Ticket size should be a constant, but it's still good to check just in case.
        title_size += len(self.tmd.dump())
        title_size += len(self.ticket.dump())
        # For contents, get their sizes from the content records, because they store the intended sizes of the decrypted
        # contents, which are usually different from the encrypted sizes.
        for record in self.content.content_records:
            if record.content_type == 32769:
                if absolute:
                    title_size += record.content_size
            else:
                title_size += record.content_size
        return title_size

    def get_title_size_blocks(self, absolute=False) -> int:
        """
        Gets the installed size of the title, including the TMD and Ticket, in the Wii's displayed "blocks" format. The
        "absolute" option determines whether shared content sizes should be included in the total size or not. This
        option defaults to False.

        1 Wii block is equal to 128KiB, and if any amount of a block is used, the entire block is considered used.

        Parameters
        ----------
        absolute : bool, optional
            Whether shared contents should be included in the total size or not. Defaults to False.

        Returns
        -------
        int
            The installed size of the title, in blocks.
        """
        title_size_bytes = self.get_title_size(absolute)
        blocks = math.ceil(title_size_bytes / 131072)
        return blocks

    def add_enc_content(self, enc_content: bytes, cid: int, index: int, content_type: int, content_size: int,
                        content_hash: bytes) -> None:
        """
        Adds a new encrypted content to the ContentRegion, and adds the provided Content ID, index, content type,
        content size, and content hash to a new record in the ContentRecord list.

        Parameters
        ----------
        enc_content : bytes
            The new encrypted content to add.
        cid : int
            The Content ID to assign the new content in the content record.
        index : int
            The index used when encrypting the new content.
        content_type : int
            The type of the new content.
        content_size : int
            The size of the new encrypted content when decrypted.
        content_hash : bytes
            The hash of the new encrypted content when decrypted.
        """
        # Add the encrypted content.
        self.content.add_enc_content(enc_content, cid, index, content_type, content_size, content_hash)
        # Update the TMD to match.
        self.tmd.content_records = self.content.content_records

    def add_content(self, dec_content: bytes, cid: int, content_type: int) -> None:
        """
        Adds a new decrypted content to the end of the ContentRegion, and adds the provided Content ID, content type,
        content size, and content hash to a new record in the ContentRecord list. The index will be automatically
        assigned by incrementing the current highest index in the records.

        This first gets the content hash and size from the provided data, and then encrypts the content with the
        Title Key before adding it to the ContentRegion.

        Parameters
        ----------
        dec_content : bytes
            The new decrypted content to add.
        cid : int
            The Content ID to assign the new content in the content record.
        content_type : int
            The type of the new content.
        """
        # Add the decrypted content.
        self.content.add_content(dec_content, cid, content_type, self.ticket.get_title_key())
        # Update the TMD to match.
        self.tmd.content_records = self.content.content_records

    def set_enc_content(self, enc_content: bytes, index: int, content_size: int, content_hash: bytes, cid: int = None,
                        content_type: int = None) -> None:
        """
        Sets the content at the provided index to the provided new encrypted content. The provided hash and content size
        are set in the corresponding content record. A new Content ID or content type can also be specified, but if it
        isn't then the current values are preserved.

        This also updates the content records in the TMD after the content is set.

        Parameters
        ----------
        enc_content : bytes
            The new encrypted content to set.
        index : int
            The index to place the new content at.
        content_size : int
            The size of the new encrypted content when decrypted.
        content_hash : bytes
            The hash of the new encrypted content when decrypted.
        cid : int
            The Content ID to assign the new content in the content record.
        content_type : int
            The type of the new content.
        """
        # Set the encrypted content.
        self.content.set_enc_content(enc_content, index, content_size, content_hash, cid, content_type)
        # Update the TMD to match.
        self.tmd.content_records = self.content.content_records

    def set_content(self, dec_content: bytes, index: int, cid: int = None, content_type: int = None) -> None:
        """
        Sets the content at the provided index to the provided new decrypted content. The hash and content size of this
        content will be generated and then set in the corresponding content record. A new Content ID or content type can
        also be specified, but if it isn't then the current values are preserved.

        This also updates the content records in the TMD after the content is set.

        Parameters
        ----------
        dec_content : bytes
            The new decrypted content to set.
        index : int
            The index to place the new content at.
        cid : int, optional
            The Content ID to assign the new content in the content record.
        content_type : int, optional
            The type of the new content.
        """
        # Set the decrypted content.
        self.content.set_content(dec_content, index, self.ticket.get_title_key(), cid, content_type)
        # Update the TMD to match.
        self.tmd.content_records = self.content.content_records

    def load_content(self, dec_content: bytes, index: int) -> None:
        """
        Loads the provided decrypted content into the ContentRegion at the specified index, but first checks to make
        sure that it matches the corresponding record. This content will then be encrypted using the title's Title Key
        before being loaded.

        Parameters
        ----------
        dec_content : bytes
            The decrypted content to load.
        index : int
            The index to load the content at.
        """
        # Load the decrypted content.
        self.content.load_content(dec_content, index, self.ticket.get_title_key())

    def fakesign(self) -> None:
        """
        Fakesigns this Title for the trucha bug.

        This is done by brute-forcing a TMD and Ticket body hash starting with 00, causing it to pass signature
        verification on older IOS versions that incorrectly check the hash using strcmp() instead of memcmp(). The TMD
        and Ticket signatures will also be erased and replaced with all NULL bytes.

        This modifies the TMD and Ticket objects that are part of this Title in place. You will need to call this method
        after any changes to the TMD or Ticket, and before dumping the Title object into a WAD to ensure that the WAD
        is properly fakesigned.
        """
        self.tmd.num_contents = self.content.num_contents  # This needs to be updated in case it was changed
        self.tmd.fakesign()
        self.ticket.fakesign()

    def get_is_fakesigned(self):
        """
        Checks the Title object to see if it is currently fakesigned. This ensures that both the TMD and Ticket are
        fakesigned. For a description of fakesigning, refer to the fakesign() method.

        Returns
        -------
        bool:
            True if the Title is fakesigned, False otherwise.

        See Also
        --------
        libWiiPy.title.title.Title.fakesign()
        """
        if self.tmd.get_is_fakesigned and self.ticket.get_is_fakesigned():
            return True
        else:
            return False

    def get_is_signed(self) -> bool:
        """
        Uses the certificate chain to verify whether the Title object contains a properly signed title or not. This
        verifies both the TMD and Ticket, and if either one fails verification then the title is not considered valid.

        This will validate the entire certificate chain. If any part of the chain doesn't match the other pieces, then
        this method will raise an exception.

        Returns
        -------
        bool
            Whether the title is properly signed or not.

        See Also
        --------
        libWiiPy.title.cert
        """
        # The entire chain needs to be verified, so start with the CA cert and work our way down. If anything fails
        # along the way, future steps don't matter so exit the descending if's and return False.
        try:
            if _verify_ca_cert(self.cert_chain.ca_cert) is True:
                if _verify_cert_sig(self.cert_chain.ca_cert, self.cert_chain.tmd_cert) is True:
                    if _verify_tmd_sig(self.cert_chain.tmd_cert, self.tmd) is True:
                        if _verify_cert_sig(self.cert_chain.ca_cert, self.cert_chain.ticket_cert) is True:
                            if _verify_ticket_sig(self.cert_chain.ticket_cert, self.ticket) is True:
                                return True
        except ValueError:
            raise ValueError("This title's certificate chain is not valid, or does not match the signature type of "
                             "the TMD/Ticket.")
        return False



# "archive/ash.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# This code in particular is a direct translation of "ash-dec" from ASH0-tools. ASH0-tools is written by Garhoogin and
# co-authored by NinjaCheetah.
# https://github.com/NinjaCheetah/ASH0-tools
#
# See <link pending> for details about the ASH compression format.

import io
from dataclasses import dataclass as _dataclass


@_dataclass
class _ASHBitReader:
    """
    An _ASHBitReader class used to parse individual words in an ASH file. Private class used by the ASH module.

    Attributes
    ----------
    src_data : list[int]
        The entire data of the ASH file being parsed, as a list of integers for each byte.
    size : int
        The size of the ASH file.
    src_pos : int
        The position in the src_data list currently being accessed.
    word : int
        The word currently being decompressed.
    bit_capacity : int
    tree_type : str
        What tree this bit reader is being used with. Used exclusively for debugging, as this value is only used in
        error messages.
    """
    src_data: list[int]
    size: int
    src_pos: int
    word: int
    bit_capacity: int
    tree_type: str


def _ash_bit_reader_feed_word(bit_reader: _ASHBitReader):
    # Ensure that there's enough data to read en entire word, then if there is, read one.
    if not bit_reader.src_pos + 4 <= bit_reader.size:
        print(bit_reader.src_pos)
        raise ValueError("Invalid ASH data! Cannot decompress.")
    bit_reader.word = int.from_bytes(bit_reader.src_data[bit_reader.src_pos:bit_reader.src_pos + 4], 'big')
    bit_reader.bit_capacity = 0
    bit_reader.src_pos += 4


def _ash_bit_reader_init(bit_reader: _ASHBitReader, src: list[int], size: int, start_pos: int):
    # Load data into a bit reader, then have it read its first word.
    bit_reader.src_data = src
    bit_reader.size = size
    bit_reader.src_pos = start_pos
    _ash_bit_reader_feed_word(bit_reader)


def _ash_bit_reader_read_bit(bit_reader: _ASHBitReader):
    # Reads the starting bit of the current word in the provided bit reader. If the capacity is at 31, then we've
    # shifted through the entire word, so a new one should be fed. If not, increase the capacity by one and shift the
    # current word left.
    bit = bit_reader.word >> 31
    if bit_reader.bit_capacity == 31:
        _ash_bit_reader_feed_word(bit_reader)
    else:
        bit_reader.bit_capacity += 1
        bit_reader.word = (bit_reader.word << 1) & 0xFFFFFFFF  # This simulates a 32-bit integer.

    return bit


def _ash_bit_reader_read_bits(bit_reader: _ASHBitReader, num_bits: int):
    # Reads a series of bytes from the current word in the supplied bit reader.
    bits: int
    next_bit = bit_reader.bit_capacity + num_bits

    if next_bit <= 32:
        bits = bit_reader.word >> (32 - num_bits)
        if next_bit != 32:
            bit_reader.word = (bit_reader.word << num_bits) & 0xFFFFFFFF  # This simulates a 32-bit integer (again).
            bit_reader.bit_capacity += num_bits
        else:
            _ash_bit_reader_feed_word(bit_reader)
    else:
        bits = bit_reader.word >> (32 - num_bits)
        _ash_bit_reader_feed_word(bit_reader)
        bits |= (bit_reader.word >> (64 - next_bit))
        bit_reader.word = (bit_reader.word << (next_bit - 32)) & 0xFFFFFFFF  # Simulate 32-bit int.
        bit_reader.bit_capacity = next_bit - 32

    return bits


def _ash_read_tree(bit_reader: _ASHBitReader, width: int, left_tree: [int], right_tree: [int]):
    # Read either the symbol or distance tree from the ASH file, and return the root of that tree.
    work = [0] * (2 * (1 << width))
    work_pos = 0

    r23 = 1 << width
    tree_root = 0
    num_nodes = 0

    while True:
        if _ash_bit_reader_read_bit(bit_reader) != 0:
            work[work_pos] = (r23 | 0x80000000)
            work_pos += 1
            work[work_pos] = (r23 | 0x40000000)
            work_pos += 1
            num_nodes += 2
            r23 += 1
        else:
            tree_root = _ash_bit_reader_read_bits(bit_reader, width)
            while True:
                work_pos -= 1
                node_value = work[work_pos]
                idx = node_value & 0x3FFFFFFF
                num_nodes -= 1
                try:
                    if node_value & 0x80000000:
                        right_tree[idx] = tree_root
                        tree_root = idx
                    else:
                        left_tree[idx] = tree_root
                        break
                except IndexError:
                    raise ValueError("Decompression failed while reading " + bit_reader.tree_type + " tree! Incorrect "
                                     "leaf width may have been used. Try using a different number of bits for the " +
                                     bit_reader.tree_type + " tree leaves.")
                # Simulate a do-while loop.
                if num_nodes == 0:
                    break
        # Also a do-while.
        if num_nodes == 0:
            break

    return tree_root


def _decompress_ash(input_data: list[int], size: int, sym_bits: int, dist_bits: int):
    # Get the size of the decompressed data by reading the second 4 bytes of the file and masking the first one out.
    decompressed_size = int.from_bytes(input_data[0x4:0x8]) & 0x00FFFFFF
    # Array of decompressed data and the position in that array that we're at. Mimics the memory pointer from the
    # original C source.
    out_buffer = [0] * decompressed_size
    out_buffer_pos = 0
    # Create two empty bit readers, and then initialize them at two different positions for the two trees.
    bit_reader1 = _ASHBitReader([0], 0, 0, 0, 0, "distance")
    _ash_bit_reader_init(bit_reader1, input_data, size, int.from_bytes(input_data[0x8:0xC], byteorder='big'))
    bit_reader2 = _ASHBitReader([0], 0, 0, 0, 0, "symbol")
    _ash_bit_reader_init(bit_reader2, input_data, size, 0xC)
    # Calculate the max for the symbol and distance trees based on the bit lengths that were passed. Then, allocate the
    # arrays for all the trees based on that maximum.
    sym_max = 1 << sym_bits
    dist_max = 1 << dist_bits
    sym_left_tree = [0] * (2 * sym_max - 1)
    sym_right_tree = [0] * (2 * sym_max - 1)
    dist_left_tree = [0] * (2 * dist_max - 1)
    dist_right_tree = [0] * (2 * dist_max - 1)
    # Read the trees to find the symbol and distance tree roots.
    sym_root = _ash_read_tree(bit_reader2, sym_bits, sym_left_tree, sym_right_tree)
    dist_root = _ash_read_tree(bit_reader1, dist_bits, dist_left_tree, dist_right_tree)
    # Main decompression loop.
    while True:
        sym = sym_root
        while sym >= sym_max:
            if _ash_bit_reader_read_bit(bit_reader2) != 0:
                sym = sym_right_tree[sym]
            else:
                sym = sym_left_tree[sym]
        if sym < 0x100:
            out_buffer[out_buffer_pos] = sym
            out_buffer_pos += 1
            decompressed_size -= 1
        else:
            dist_sym = dist_root
            while dist_sym >= dist_max:
                if _ash_bit_reader_read_bit(bit_reader1) != 0:
                    dist_sym = dist_right_tree[dist_sym]
                else:
                    dist_sym = dist_left_tree[dist_sym]
            copy_len = (sym - 0x100) + 3
            srcp_pos = out_buffer_pos - dist_sym - 1
            # Check to make sure we aren't going to exceed the specified decompressed size.
            if not copy_len <= decompressed_size:
                raise ValueError("Invalid ASH data! Cannot decompress.")

            decompressed_size -= copy_len
            while copy_len > 0:
                out_buffer[out_buffer_pos] = out_buffer[srcp_pos]
                out_buffer_pos += 1
                srcp_pos += 1
                copy_len -= 1
        # Simulate a do-while loop.
        if decompressed_size == 0:
            break

    return out_buffer


def decompress_ash(ash_data: bytes, sym_tree_bits: int = 9, dist_tree_bits: int = 11) -> bytes:
    """
    Decompresses the data of an ASH file and returns the decompressed data.

    With the default parameters, this function can decompress ASH files found in the files of the Wii Menu and Animal
    Crossing: City Folk. Some ASH files, notably the ones found in the WiiWare title My Pokémon Ranch, require setting
    dist_tree_bits to 15 instead for a successful decompression. If an ASH file is failing to decompress with the
    default options, trying a dist_tree_bits value of 15 will likely fix it. No other leaf sizes are known to exist,
    however they might be out there.

    Parameters
    ----------
    ash_data : bytes
        The data for the ASH file to decompress.
    sym_tree_bits : int, option
        Number of bits for each leaf in the symbol tree. Defaults to 9.
    dist_tree_bits : int, option
        Number of bits for each leaf in the distance tree. Defaults to 11.
    """
    # Check the magic number to make sure this is an ASH file.
    with io.BytesIO(ash_data) as ash_data2:
        ash_magic = ash_data2.read(4)
        if ash_magic != b'\x41\x53\x48\x30':
            raise TypeError("This is not a valid ASH file!")
    # Begin decompression. Convert the compressed data to an array of ints for processing, then convert the returned
    # decompressed data back into bytes to return it.
    ash_size = len(ash_data)
    ash_data_int = [byte for byte in ash_data]
    decompressed_data = _decompress_ash(ash_data_int, ash_size, sym_tree_bits, dist_tree_bits)
    decompressed_data_bin = bytes(decompressed_data)

    return decompressed_data_bin
